import com.couchbase.client.java.Cluster;
import com.couchbase.client.java.ReactiveCollection;
import com.couchbase.client.java.json.JsonArray;
import com.couchbase.client.java.json.JsonObject;
import com.couchbase.client.java.kv.MutationResult;
import com.couchbase.client.java.kv.PersistTo;
import com.couchbase.client.java.kv.ReplicateTo;
import com.couchbase.client.java.kv.UpsertOptions;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.stream.Collectors;

public class Couchbase3xBulkInsert {
    public static void main(String[] args) {
        System.out.println("Couchbase3xBulkInsert");

        Cluster cluster = Cluster.connect("127.0.0.1", "data1", "password");
        ReactiveCollection reactiveCollection = cluster.bucket("data1").defaultCollection().reactive();

        // Create a JSON Document
        JsonObject arthur = JsonObject.create()
                .put("name", "Arthur")
                .put("email", "kingarthur@couchbase.com")
                .put("interests", JsonArray.from("Holy Grail", "African Swallows"));

        JsonObject arthurJunior = JsonObject.create()
                .put("name", "ArthurJr")
                .put("email", "kingarthurjr@couchbase.com")
                .put("interests", JsonArray.from("Holy Grail", "African Swallows"));

        JsonObject arthurBaby = JsonObject.create()
                .put("name", "ArthurBaby")
                .put("email", "kingarthurbaby@couchbase.com")
                .put("interests", JsonArray.from("Holy Grail", "African Swallows"));

        // Create an array of things that should fail
        Map<String, JsonObject> entities = new HashMap();
        entities.put("U:king_arthur", arthur);
        entities.put("U:king_arthur_jr", arthurJunior);
        entities.put("U:king_arthur_baby"
                        + "12345678901234567890123456789012345678901234567890"
                        + "12345678901234567890123456789012345678901234567890"
                        + "12345678901234567890123456789012345678901234567890"
                        + "12345678901234567890123456789012345678901234567890"
                        + "12345678901234567890123456789012345678901234567890",
                arthurBaby);

        entities.put(null, arthur);

        UpsertOptions emptyUpsertOptions = UpsertOptions.upsertOptions();

        UpsertOptions failUpsertOptions = UpsertOptions.upsertOptions()
                .durability(PersistTo.FOUR, ReplicateTo.THREE);

        // Map to collect exceptions during async inserts, use ConcurrentHashMap since it needs to be thread-safe
        Map<String, Throwable> upsertProblems = new ConcurrentHashMap();

        List<MutationResult> successResults = Collections.synchronizedList(new ArrayList());

        Flux
                .fromIterable(entities.keySet())
                .flatMap(key -> reactiveCollection.upsert(
                        key, entities.get(key), !key.contains("jr") ? emptyUpsertOptions : failUpsertOptions)
                        .onErrorResume(e -> {
                    System.out.println(e.getClass().toString() + " " + e.getMessage()
                            + ": Could not upsert document with Id: " + key);
                    upsertProblems.put(key, e);

                    // Note MutationResult can't be null
                    // so we blockLast() to ensure we get what we need.
                    return Mono.empty();
                }))
                .doOnNext(successResults::add)
                .blockLast();

        // Aggregate & return results
        // In this context, CANCELLED (FAILED_DOCUMENT_STATUS) may mean that we were unable to create a Document
        // to upload to Couchbase; it does not necessarily mean that the request to Couchbase was cancelled. It's
        // possible to differentiate this error from others by inspecting the error message.
        Map<String, Boolean> results = entities.keySet().stream()
                .collect(Collectors.toMap(
                        Function.identity(),
                        k -> !upsertProblems.containsKey(k)));

        results.forEach((k, v) -> {
            System.out.println("Key: " + k + " Success: " + v);
        });
    }
}
